#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''A simple TiMBL command but splits the test file into parts and runs them simultaneously on multiple CPUs

Copyright (c) 2013 CLiPS. All rights reserved.

# License: GNU General Public License, see http://www.clips.ua.ac.be/~vincent/scripts/LICENSE.txt
'''
__version__='1.1.1'
__date__='May 2013'
__author__='Vincent Van Asch'
import os, sys, getopt, time, re
import multiprocessing, tempfile, subprocess, signal, threading

# The default TiMBL location
TIMBL = '~/bin/timbl'

def loginfo(s): print >>sys.stderr, '%s: %s' %(time.strftime('%d/%m/%Y %H:%M:%S'), s)

def fread(fname, sep=None):
    with open(os.path.abspath(os.path.expanduser(fname)), 'rU') as f:
        for l in f:
            line = l.strip()
            if line: yield line.split(sep)

parts = re.compile('(\d+)')
def shift_options(options):
    # All indices +1
    s=''
    for ch in parts.split(options):
        try:
            ch = int(ch)
        except Exception:
            s+=ch
        else:
            s = s + '%d' %(ch+1) 
            
    # Add the ignore
    if 'I' in s:
        s = s.replace('I', 'I1,')
    else:
        s+=':I1'
            
    
    return s

def split_options(options):
    '''Creates the shift for the indices.
    Returns a list'''
    if not options: return ['-mO:I1']
    
    if '-m' not in options: return options.split()
    
    # Search -m
    out=[]
    i=0
    options = options.split()
    while i < len(options):
        option = options[i]
        if option == '-m':
            out.append(option)
            # we need to change the next one
            out.append(shift_options(options[i+1]))
        
            i+=1
        elif option.startswith('-m'):
            out.append('-m')
            out.append(shift_options(option[2:]))
            
        elif '-m' in option:
            raise ValueError('Unexpected option: %s' %option)
        else:
            out.append(option)
            
        i+=1

    return out
        

def safe_count(fname):
    '''Returns the number of lines in a file'''
    try:
        return len(list(fread(fname)))
    except IOError:
        return -1

def reporter(fname, ninstances):
    '''Check and report the number of lines in a file'''
    nlines = safe_count(fname)
    pp=-1
    while nlines < ninstances:
        nlines = safe_count(fname)

        p = int(round(10.0*nlines/ninstances))
        
        if p == 0: continue
        if p == 10: break
        
        if p != pp:
            loginfo('Approximately %2d%% processed...' %(10*p))
            pp = p
        time.sleep(2)

        
def runcommand(cmd):
    proc = subprocess.Popen(cmd, stderr = subprocess.PIPE, stdout=subprocess.PIPE)
    out, err = proc.communicate()
    
    
def subrun(args):
    cmd, verbose, ninstances = args
    
    if verbose:
        # If we need to report on the progress, we have to split into threads
        out, err = 'out', 'err'
        
        # Init
        reporterthread = threading.Thread(name='reporter', target=reporter, args=(cmd[-1], ninstances))
        timblthread = threading.Thread(name='timbl', target=runcommand, args=(cmd,))
        
        # Start and join
        reporterthread.start()
        timblthread.start()
        reporterthread.join()
        timblthread.join()
        
    else:
        proc = subprocess.Popen(cmd, stderr = subprocess.PIPE, stdout=subprocess.PIPE)
        out, err = proc.communicate()
    return out, err

def init_worker():
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def run(trainingfile, testfile, cpu=None, timbl=TIMBL, verbose=True, sep=None, options='', output=None):
    '''Run the TiMBL experiment.
    
    trainingfile: a timbl training file
    testfile: a timbl test file
    
    cpu: number of CPUs to use (None is all)
    timbl: path to TiMBL binary
    sep: feature separator (important to specify if this is not whitespace)
    options: the Timbl options as a string
    output: the name of the output file to create (default: testfile.out)
    
    Returns the name of the output file
    '''
    # The number of cpu's to use
    N=multiprocessing.cpu_count()
    if cpu:
        if cpu > N:
            loginfo('Can only use %d CPUs' %N)
        else:
            N = cpu

    # Paths
    trainingfile = os.path.abspath(os.path.expanduser(trainingfile))
    testfile = os.path.abspath(os.path.expanduser(testfile))
    timbl = os.path.abspath(os.path.expanduser(timbl))
    if output is None:
        output = testfile+'.out'
    elif output.strip():
        output = os.path.abspath(os.path.expanduser(output))
    else:
        raise ValueError('Empty output path. See option -O.')

    if verbose:
        loginfo('TRAINING: %s' %trainingfile)
        loginfo('TEST    : %s' %testfile)
        loginfo('Using %d CPUs from %d' %(N, multiprocessing.cpu_count()))
        if sep: loginfo('FEATURE SEPARATOR: %s' %sep)
        loginfo('TiMBL options: %s' %options)
        loginfo('Writing output to: %s' %output)
        
    # Feature separator
    fs = sep
    if sep is None: fs=' '
        
    # Split and prepare the indices in the TiMBL options
    options = split_options(options)
        
    try:
        # Prepare the training file
        id=0
        fd, train = tempfile.mkstemp(prefix='simtimbl.train.')
        for line in fread(trainingfile, sep=sep):
            os.write(fd, '%s\n' %(fs.join(['tr'+str(id)]+line)) )
            id+=1
        os.close(fd)
            
        # Create the command (except the testfiles and outputfiles)
        cmd = [timbl, '-f', train] + options
            
        # Split the test file in N parts
        # Open tempfiles
        testfiles = []
        for i in range(N):
            testfiles.append( tempfile.mkstemp(prefix='simtimbl.', suffix=('.%d' %i)) )
            
        # Split instances over files
        id=0
        for i, instance in enumerate(fread(testfile, sep=sep)):
            index = i%N
            os.write( testfiles[index][0], '%s\n' %(fs.join([str(id)]+instance)))
            id+=1
        ninstances = id
        
        # Close files
        for fd, name in testfiles:
            os.close(fd)
            
        if verbose: loginfo('Split testfile in %d parts' %N)
            
        # Create all commands
        cmds=[]
        tasks=[]
        outnames=[]
        report=verbose
        for fd, name in testfiles:
            outname = name + '.out'
            
            nlines=0
            if report:
                nlines = safe_count(name)
            
            tasks.append( (cmd + ['-t', name, '-o', outname], report, nlines) )
            #print ' '.join(tasks[-1])
            outnames.append(outname)
            report=False
            
        # Processing
        # Create the processing pool
        pool = multiprocessing.Pool(N, init_worker)
        results =  pool.imap_unordered(subrun, tasks)
        
        if verbose: loginfo('Starting parallel TiMBL experiment')
        
        # Wait until finshed
        count=1
        try:
            for part in results:
                #if part[1]: print part[1]
                if verbose: loginfo('Finished part %d of %d' %(count, N))
                count+=1
        except OSError:
            loginfo('ERROR: Most probably "%s" is not a working TiMBL binary. See option -b.' %timbl)
            sys.exit(1)
            
        # Recreate testfile
        if verbose: loginfo('Reconstitute test file from parts')
        
        # Get all readers
        readers=[]
        for outname in outnames:
            readers.append( fread(outname, sep=sep) )
            
        # Loop over all instances
        count=0
        with open(output, 'w') as out:
            while 1:
                try:
                    instance = readers[count%N].next()
                except IOError:
                    loginfo('ERROR: the TiMBL output file does not exist. Most probable the option string -o contains an error.')
                    sys.exit(1)
                except StopIteration:
                    break
                    
                id = int(instance[0])
                if id != count:
                    raise ValueError('bad aligment: %d <> %d' %(id, count))
                
                out.write('%s\n' %(fs.join(instance[1:])) )
                count+=1
                
        if ninstances != count:
            raise ValueError('Expected %d instances, got %s' %(ninstances, count))
        
    finally:
        # Clean up
        for outname in outnames:
            try:
                os.remove(outname)
            except Exception:
                pass
        for fd, name in testfiles:
            try:
                os.remove(name)
            except Exception:
                pass
        try:
            os.remove(train)
        except Exception:
            pass
        
    if verbose: loginfo('Output written to %s' %output)
    return output
        

def _usage():
    print >>sys.stderr, '''Runs a single TiMBL experiment using all CPUs (version %s)
    
USAGE
    $ python simtimbl.py [-F sep] [-b path] [-c n] [-v] [-O path] [-o options] trainingfile testfile
    
    files: TiMBL instance files
    
OPTIONS
    -b path   : the path to a working TiMBL binary (default: %s)
    -c n      : Use n CPUs (default: all)
    -F sep    : the feature separator. Important to supply if this is not whitespace
                (default: whitespace)
    -O path   : path to write the output file to (default: testfile.out)
    -o options: a string with all TiMBL options that should be set
                between quotes, example: "-mM -dIL -k6"
    -v        : verbose

NOTE
    The output file is exactly as the output file of the "pure" TiMBL
    command would look like.

    During processing an id is added at the beginning of the instances. This means
    that gainratio and other feature index based information is shifted one position
    upwards. If you only use the output file, this behavior is unnoticeable.

SHALLOW COMPARISON
    simtimbl.py works differently than Dimbl. simtimbl.py does not split up
    the training corpus. This means that the output of simtimbl.py is
    exactly the same as from the equivalent TiMBL-command. This is
    not the case for Dimbl.
    
    A quick test using a training corpus of 48,868 instances and a
    test corpus of 79,994 instances resulted in the following timings,
    relative to the equivalent TiMBL-command:
        4CPU's:
            - TiMBL       : duration 1.00
            - simtimbl.py : duration 0.31
            - Dimbl       : duration 0.36
        8CPU's
            -simtimbl.py  : duration 0.15

REFERENCE
    - See for TiMBL: http://ilk.uvt.nl/timbl
    - You can use confusionmatrix.py to get an evaluation of the output file:
       http://www.clips.ua.ac.be/~vincent/software.html#confusionmatrix

%s, %s
''' %(__version__, TIMBL, __author__, __date__)
    
    
    
    
if __name__ == '__main__':        
    try:
        opts,args=getopt.getopt(sys.argv[1:],'hc:b:vF:o:O:', ['help'])
    except getopt.GetoptError:
        # print help information and exit:
        _usage()
        sys.exit(2)

    cpu=None
    timbl=TIMBL 
    verbose=False
    sep=None
    options=''
    output=None
    for o, a in opts:
        if o in ('-h', '--help'):
            _usage()
            sys.exit()
        if o == '-c':
            cpu = int(a)
        if o == '-b':
            timbl = os.path.abspath(os.path.expanduser(a))
        if o =='-v':
            verbose=True
        if o =='-F':
            sep=a
        if o == '-o':
            options = a
        if o == '-O':
            output = os.path.abspath(os.path.expanduser(a))

    if len(args) != 2:
        _usage()
        sys.exit(1)

    trainingfile, testfile = args
    run(trainingfile, testfile, cpu=cpu, timbl=timbl, verbose=verbose, sep=sep, options=options, output=output)
